#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import asyncio

from aioutils import Group, Pool

def test_group():
    timespan = 0.1
    @asyncio.coroutine
    def f(i):
        t0 = time.time()
        yield from asyncio.sleep(timespan)
        t = time.time() - t0
        print('finish {}, seconds={:4.2f}'.format(i, t))

    print('testing group')
    t0 = time.time()
    g = Group()
    for i in range(9):
        g.spawn(f(i))
    g.join()
    print('total time: {:4.2f}'.format(time.time() - t0))
    assert timespan < time.time() - t0 < timespan * 1.1


def test_pool():
    timespan = 0.1

    @asyncio.coroutine
    def f(i):
        t0 = time.time()
        yield from asyncio.sleep(timespan)
        t = time.time() - t0
        print('finish {}, seconds={:4.2f}'.format(i, t))

    print('testing pool')
    t0 = time.time()
    p = Pool(3)
    for i in range(9):
        p.spawn(f(i))
    p.join()
    print('total time: {:4.2f}'.format(time.time() - t0))
    assert timespan * 3 < time.time() - t0 < timespan * 3 * 1.1

if __name__ == '__main__':
    test_group()
    test_pool()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Bag and OrderedBag

This is old implementation for yielding helper
You should use Yielder and OrderedYielder instead
"""
import time
import queue
import inspect
import asyncio
import functools
import threading

from .pool import Group


class Bag(object):

    """ Create generator function with coroutines

    A Bag is just a Group, a queue, and a background thread running event loop.
    Coroutines are spawned in a schedule method, and enqueues result to the
    queue, main thread then yield items from queue.

    Usage::

    def gen_func():
        b = Bag()
        @asyncio.coroutine
        def coro(arg):
            ...
            b.put(...)

        def schedule():
            ...
            for ...:
                b.spawn(coro(arg))
            ...
            b.join()

        b.schedule(schedule)
        yield from b.yielder()
    """

    def __init__(self, group=None):
        if group is not None:
            self.g = group
            self.loop = self.g.loop
        else:
            try:
                self.loop = asyncio.get_event_loop()
                if self.loop.is_running():
                    raise NotImplementedError("Cannot use aioutils in "
                                            "asynchroneous environment")
            except:
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)
            self.g = Group(loop=self.loop)
        self.q = queue.Queue()
        self.t = None

    def spawn(self, coro):
        return self.g.spawn(coro)

    def join(self):
        return self.g.join()

    def put(self, item):
        self.q.put(item)

    def schedule(self, schedule):
        def schedule_wrapper(loop):
            asyncio.set_event_loop(loop)
            schedule()

        self.t = threading.Thread(target=schedule_wrapper, args=(self.loop,))
        self.t.start()

    def yielder(self):
        while (self.t and self.t.is_alive()) or (self.q.qsize() > 0):
            try:
                yield self.q.get_nowait()
            except:
                time.sleep(0.1)


class OrderedBag(Bag):

    """ A Bag that ensures ordering """

    def __init__(self, *args):
        super(OrderedBag, self).__init__(*args)
        self.q = queue.PriorityQueue()
        self.order = 0

    def spawn(self, coro):
        self.order += 1

        def _coro_wrapper(order):
            yield from coro
        return self.g.spawn(functools.partial(_coro_wrapper, self.order)())

    def put(self, item):
        order = self._get_coro_order()
        self.q.put((order, item))

    def _get_coro_order(self):
        """ Get the coro (and its order) of self """
        stack = inspect.stack()
        for frame, module, line, function, context, index in stack:
            if function == '_coro_wrapper':
                return frame.f_locals['order']

    def yielder(self):
        next_order = 1
        while (self.t and self.t.is_alive()) or (self.q.qsize() > 0):
            try:
                # peek item, the smallest of heap
                order, item = self.q.queue[0]
                if order != next_order:
                    time.sleep(0.03)
                else:
                    order, item = self.q.get_nowait()
                    next_order += 1
                    yield item
            except IndexError:
                time.sleep(0.03)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import random
import inspect
import asyncio
from aioutils import Yielder, OrderedYielder, yielding, ordered_yielding

from nose.tools import raises

@asyncio.coroutine
def f(c):
    yield from asyncio.sleep(random.random()*.1)
    return c


def test_yielder():
    def gen_func():
        y = Yielder()
        @asyncio.coroutine
        def f1(i):
            yield from asyncio.sleep(random.random()*0.1)
            for j in [random.randint(0, 10) for _ in range(i)]:
                y.spawn(f2(j))

        @asyncio.coroutine
        def f2(j):
            yield from asyncio.sleep(random.random()*0.1)
            return j + random.random()

        for i in range(5):
            y.spawn(f1(i))

        y.put(3)

        for i in range(5, 10):
            y.spawn(f1(i))

        yield from y.yielding()

    g = gen_func()
    assert inspect.isgenerator(g)
    t0 = time.time()
    gs = list(g)
    assert len(gs) == 46
    assert min(gs) > 0 and max(gs) < 11
    assert time.time() - t0 < 0.2 * 1.1


def test_ordered_yielder():
    chars = 'abcdefg'

    def gen_func():
        y = OrderedYielder()
        for c in chars[:3]:
            y.spawn(f(c))

        y.put('z')

        for c in chars[3:]:
            y.spawn(f(c))

        yield from y.yielding()

    g = gen_func()
    assert inspect.isgenerator(g)
    t0 = time.time()
    gs = list(g)
    assert ''.join(gs) == chars[:3] + 'z' + chars[3:]
    assert time.time() - t0 < 0.1 * 1.1


def test_yielding():
    chars = 'abcdefg'
    def gen_func():
        with yielding() as y:
            for c in chars:
                y.spawn(f(c))
            yield from y

    def gen_func2():
        with ordered_yielding() as y:
            for c in chars:
                y.spawn(f(c))
            yield from y

    g = gen_func()
    assert inspect.isgenerator(g)
    t0 = time.time()
    gs = list(g)
    assert set(gs) == set(chars)
    assert time.time() - t0 < 0.1 * 1.1

    assert ''.join(list(gen_func2())) == chars


def test_yielder_with_pool_size():
    chars = 'abcdefgh'
    def gen_func():
        with yielding(2) as y:
            for c in chars:
                y.spawn(f(c))
            yield from y


    # I don't know how to assert correctness, >.<
    assert set(gen_func())  == set(chars)


def test_empty_yielder():
    def gen_func():
        with yielding() as y:
            y.spawn(asyncio.sleep(0.01))
            yield from y

    assert list(gen_func()) == []


def test_two_level_ordered_yielding():
    chars = 'abcdefg'
    def gen_func():
        with ordered_yielding() as y:
            @asyncio.coroutine
            def g(i):
                for c in chars[:i]:
                    y.spawn(f(c))

            for i in range(3, 5):
                y.spawn(g(i))

            yield from y

    assert ''.join(list(gen_func())) == 'abcabcd'


def test_break_from_yielding():
    @asyncio.coroutine
    def g(c):
        yield from asyncio.sleep(random.random()*.1)
        return c
    y = Yielder()

    def gen_func():
        for c in 'abcdefg':
            y.spawn(g(c))
        yield from y.yielding()

    for x in gen_func():
        break

    assert y.counter == 0


@raises(ValueError)
def test_raise_from_yielding():
    @asyncio.coroutine
    def g(c):
        yield from asyncio.sleep(random.random()*.1)
        if random.random() < 0.2:
            raise ValueError
        return c

    def gen_func():
        with yielding(3) as y:
            for c in 'abcdefghijklmn':
                y.spawn(g(c))
            yield from y

    # test that raise will not cause problem
    for x in gen_func():
        pass


@raises(ValueError)
def test_raise_from_nested_yielding():
    yielder = Yielder()

    @asyncio.coroutine
    def worker(err_count=0):
        yield from asyncio.sleep(random.random()*.02)
        if err_count > 5:
            raise ValueError
        else:
            return (yield from worker(err_count+1))

    for i in range(10):
        yielder.spawn(worker())

    for i, ch in enumerate(yielder.yielding()):
        pass


if __name__ == '__main__':
    test_yielder()
    test_ordered_yielder()
    test_yielding()
    test_yielder_with_pool_size()
    test_empty_yielder()
    test_two_level_ordered_yielding()
    test_break_from_yielding()
    test_raise_from_yielding()
    test_raise_from_nested_yielding()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Bag and OrderedBag """
import heapq
import asyncio
import functools
import collections


class Yielder(object):

    """ A Bag Rewrite

    - easier to use
    - no background threading

    Each time when an item is put, we stop the main loop(!!) and yield
    """

    def __init__(self, pool_size=None):
        try:
            self.loop = asyncio.get_event_loop()
            if self.loop.is_running():
                raise NotImplementedError("Cannot use aioutils in "
                                          "asynchroneous environment")
        except:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self.sem = asyncio.Semaphore(pool_size) if pool_size else None
        self._prepare()

    def _prepare(self):
        self.counter = 0
        self.done = collections.deque()
        self.getters = collections.deque()
        self.exceptions = []
        self.tasks = []

    def spawn(self, coro):
        task = self._async_task(coro)
        task.add_done_callback(self._on_completion)
        self.counter += 1
        self.tasks.append(task)
        return task

    def _async_task(self, coro):
        if self.sem:
            @asyncio.coroutine
            def _limit_coro():
                with (yield from self.sem):
                    return (yield from coro)
            task = asyncio.async(_limit_coro())
        else:
            task = asyncio.async(coro)
        return task

    def _on_completion(self, f):
        self.counter -= 1
        f.remove_done_callback(self._on_completion)
        try:
            result = f.result()
        except Exception as e:
            if not isinstance(e, asyncio.InvalidStateError):
                self.exceptions.append(e)
            result = None
        if result is not None:
            self._put(result)
        if self.counter <= 0 and self.getters:
            getter = self.getters.popleft()
            getter.set_result(None)

    def put(self, item):
        self._put(item)

    def _put(self, item):
        if self.getters:
            getter = self.getters.popleft()
            getter.set_result(None)

        self.done.append(item)

    def _stop_loop(self, f):
        self.loop.stop()

    def _yielding(self):
        while self.counter > 0 or self.done:
            if self.done:
                yield self.done.popleft()
            else:
                getter = asyncio.Future()
                self.getters.append(getter)
                getter.add_done_callback(self._stop_loop)
                if not self.loop.is_running():
                    self.loop.run_forever()

    def yielding(self):
        try:
            for x in self._yielding():
                if isinstance(x, asyncio.Future):
                    continue
                yield x
        except GeneratorExit:
            for task in self.tasks:
                if not task.done():
                    task.cancel()
                    self.counter -= 1

        if self.exceptions:
            raise self.exceptions[0]
        self._prepare()


class OrderedYielder(Yielder):
    def __init__(self, pool_size=None):
        super(OrderedYielder, self).__init__(pool_size)
        self._prepare()

    def _prepare(self):
        super(OrderedYielder, self)._prepare()
        self.done = []
        self.order = 0
        self.yield_counter = 0

    def spawn(self, coro):
        self.order += 1
        task = self._async_task(coro)
        task.add_done_callback(
            functools.partial(self._on_completion, order=self.order))
        self.counter += 1
        self.tasks.append(task)
        return task

    def _on_completion(self, f, order):
        self.counter -= 1
        f.remove_done_callback(self._on_completion)
        try:
            result = f.result()
        except Exception as e:
            if not isinstance(e, asyncio.InvalidStateError):
                self.exceptions.append(e)
            result = None
        self._put((order, result))

    def _put(self, item, heappush=heapq.heappush):
        order, item = item
        if self.yield_counter + 1 == order or self.counter <= 0:
            if self.getters:
                getter = self.getters.popleft()
                getter.set_result(None)

        heappush(self.done, (order, item))

    def put(self, item):
        self.order += 1
        self._put((self.order, item))

    def _yielding(self, heappop=heapq.heappop):
        self.yield_counter = 1
        while self.counter > 0 or self.done:
            if self.done:
                order, item = self.done[0]
                if self.yield_counter == order:
                    _, item = heappop(self.done)
                    if item is not None:
                        yield item
                    self.yield_counter += 1
                    continue

            getter = asyncio.Future()
            self.getters.append(getter)
            getter.add_done_callback(self._stop_loop)
            if not self.loop.is_running():
                self.loop.run_forever()


class YieldingContext(object):
    def __init__(self, pool_size=None, ordered=False):
        if ordered:
            self.y = OrderedYielder(pool_size)
        else:
            self.y = Yielder(pool_size)
        self.yielding = None

    def spawn(self, coro):
        return self.y.spawn(coro)

    def put(self, item):
        return self.y.put(item)

    def __enter__(self):
        return iter(self)

    def __iter__(self):
        self.yielding = self.y.yielding()
        return self

    def __next__(self):
        return next(self.yielding)

    def __exit__(self, *args):
        pass

yielding = YieldingContext
ordered_yielding = functools.partial(YieldingContext, ordered=True)
